#ifndef _CONFIG_H
#define _CONFIG_H

#include <stdio.h>

typedef struct {
  int index;
  char flag;
  char* name;
} verbose_flag_t;

#define VFLAG_TIMERS 1
#define VFLAG_AGENTS 2
#define VFLAG_LOCS 4
#define VFLAG_LINKS 8
#define VFLAG_UPDATE 16
#define VFLAG_ADD_AGENT 32
#define VFLAG_EXTEND 64
#define VFLAG_ADD_LINK 128
#define VFLAG_DISTS 256
#define VFLAG_APROBS 512

static const verbose_flag_t VERBOSE_FLAGS[] = {
  {.index = VFLAG_TIMERS, .flag = 'T', .name = "print timers"},
  {.index = VFLAG_AGENTS, .flag = 'A', .name = "print agents"},
  {.index = VFLAG_LOCS, .flag = 'L', .name = "print locs"},  
  {.index = VFLAG_LINKS, .flag = 'X', .name = "print links"},
  {.index = VFLAG_UPDATE, .flag = 'U', .name = "print update"},
  {.index = VFLAG_ADD_AGENT, .flag = 'N', .name = "add agent"},
  {.index = VFLAG_EXTEND, .flag = 'E', .name = "extend agent"},
  {.index = VFLAG_ADD_LINK, .flag = 'I', .name = "add links"},
  {.index = VFLAG_DISTS, .flag = 'D', .name = "print distances"},
  {.index = VFLAG_APROBS, .flag = 'P', .name = "print attachment probabilities"},
};

#define DBG(FLAG, fmt, ...)                                             \
  do {									\
    if (FLAG & _config.verbose_flags)					\
      printf("[%d:%d] " fmt, _sim_iters, MYTHREAD, __VA_ARGS__);	\
  } while (0);

#define SPREAD_GEO_SEL_LSTRAT 1
#define FOCUS_GEO_SEL_LSTRAT 2 
#define SPREAD_URND_SRC_LSTRAT 3
#define FOCUS_URND_SRC_LSTRAT 4
#define CHAIN_LSTRAT 5
#define URND_FIXED_GEO_LSTRAT 6
#define GEO_SEL_FIXED_GEO_LSTRAT 7

static const int NUM_LOCS = 45000;
static const int MAX_DEG = 4000;
static const int MAX_POP = 10000;
static const int GRAPH_EVAL_FREQ = 16;
static const int NUM_AGENTS_WITH_BASE_INCOME = 40000;//100;
static const int LINK_STRATEGY = CHAIN_LSTRAT;
// this is not yet correct. If a link is removed, it can partition the graph. 
// Somehow we have to avoid that.
//static const int REMOVE_LINKS = 0;

typedef struct {
  int output_interval; 
  unsigned int rseed;  
  int max_agents;
  double pop_distr_exp;
  double av_degree;    
  double ext_cost;
  int base_income;
  double wickedness;
  shared char* snapshot_fname;
  int full_traffic;
  double intrvn_drop_ingress;
  double intrvn_drop_egress;
  double intrvn_drop_transit;
  int intrvn_large_isp_deg;
  double intrvn_active_prob;
  double recovery_rate;
  double infct_rate;
  double wkdns_to_block;
  int too_big_too_block;
  int verbose_flags;
} config_t;

void load_config(int argc, char** argv, shared config_t* config);
void print_config(shared config_t* config, char comment, FILE* f);

#endif

