#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "utils.h"
#include "config.h"
#include "asim.h"

char* get_format(char* buf, char* fmt, int length) {
  char buf2[1000];
  strcpy(buf, "");
  sprintf(buf2, "%d", length + 1);
  strcat(buf, "%-");
  strcat(buf, buf2);
  strcat(buf, fmt);
  return buf;
}

typedef struct {
  double ts[MAX_TIMERS];
} timer_info_t;


void save_loc_snapshot(FILE* f);
void restore_loc_snapshot(FILE* f);
void save_agent_snapshot(FILE* f);
void restore_agent_snapshot(FILE* f);

void print_timers() {
  char buf[1000], buf2[1000];
  shared timer_info_t* timers = upc_all_alloc(THREADS, sizeof(timer_info_t));
  upc_barrier;
  for (int i = 0; i < MAX_TIMERS; i++) timers[MYTHREAD].ts[i] = timer_read(i);
  upc_barrier;
  SERIAL {
    printf("Timers:\n");
    for (int i = 2; i < MAX_TIMERS && _timer_names[i]; i++) {
      if (strcmp(_timer_names[i], "(barriers)") != 0) {
        printf(get_format(buf, "s ", strlen(_timer_names[i])), _timer_names[i]);
      }
    }
    printf("\n");
    double tot_timers[MAX_TIMERS], max_timers[MAX_TIMERS], min_timers[MAX_TIMERS];
    bzero(max_timers, MAX_TIMERS * sizeof(double));
    bzero(tot_timers, MAX_TIMERS * sizeof(double));
    for (int j = 2; j < MAX_TIMERS && _timer_names[j]; j++) min_timers[j] = 1e12;
    for (int i = 0; i < THREADS; i++) {    
      for (int j = 2; j < MAX_TIMERS && _timer_names[j]; j++) {
        tot_timers[j] += timers[i].ts[j];
        if (timers[i].ts[j] > max_timers[j]) max_timers[j] = timers[i].ts[j];
        if (timers[i].ts[j] < min_timers[j]) min_timers[j] = timers[i].ts[j];
	/*
	if (strcmp(_timer_names[j], "(barriers)") != 0) {
	  sprintf(buf, "%.1f/%.1f", timers[i].ts[j], timers[i].ts[j + 1]);
	  printf(get_format(buf2, "s ", strlen(_timer_names[j])), buf);
	}
	*/
      }
      //      printf("\n");
    }
    for (int i = 2; i < MAX_TIMERS && _timer_names[i]; i++) {
      if (strcmp(_timer_names[i], "(barriers)") != 0) {
        sprintf(buf, "%.1f/%.1f", tot_timers[i] / THREADS, tot_timers[i + 1] / THREADS);
        printf(get_format(buf2, "s ", strlen(_timer_names[i])), buf);
      }
    }
    printf("\n");
  }
}

void print_agents() {
  char fname[100];
  sprintf(fname, "agents-%d.dat", _agents.num);
  FILE* f = fopen(fname, "w+");
  print_config(&_config, '#', f);
  mfprintf(f, 15, "# %5s", "id/0", 
	   "%8s", "age/1", 
	   "%10s", "popsh/2", 
	   "%10s", "pop/3",
	   "%8s", "deg/4", 
	   "%8s", "nlocs/5", 
	   "%7s", "ad/6", 
	   "%5s", "md/7", 
	   "%10s", "traf/8", 
	   "%10s", "bt/9", 
	   "%10s", "lbt/10", 
	   "%8s", "wck/11",  
	   "%10s", "wckhs/12", 
	   "%7s", "act/13", 
	   "%10s", "lgt/14");
  if (_agents.num < 100) mfprintf(f, 2, "  %-30s", "links", "%-30s", "locs");
  fprintf(f, "\n");
  for (int i = 0; i < _agents.num; i++) {
    shared agent_t* agent = &_agents._[i];
    char buf[10000], link_str[10000] = "", loc_str[10000] = "";
    if (_agents.num < 100) {
      // don't print more than a limited number of items
      int num_items = (agent->num_links > 15 ? 15 : agent->num_links);
      for (int j = 0; j < num_items; j++) {
	sprintf(buf, "%d,", agent->links_list[j]);;
	strcat(link_str, buf);
      }
      num_items = (agent->num_locs > 15 ? 15 : agent->num_locs);
      for (int l = 0; l < num_items; l++) {
	sprintf(buf, "%d,", agent->locs_list[l]);
	strcat(loc_str, buf);
      }
    } 
    int age = _sim_iters - agent->birth_time;
    mfprintf(f, 15, "  %5d", agent->id, 
	     "%8d", age, 
	     "%10.2e", (double)agent->pop_sh / (double)_config.max_agents, 
	     "%10.2e", (double)agent->tot_pop,
	     "%8d", agent->num_links, 
	     "%8d", agent->num_locs,
	     "%7.3f", agent->av_dist, 
	     "%5.0f", agent->max_dist, 
	     "%10.2e", (double)agent->end_user_traffic, 
	     "%10.2e", (double)agent->bad_traffic,
	     "%10.2e", (double)agent->lost_bad_traffic,
	     "%8.5f", agent->wickedness,
	     "%10.2e", (double)agent->pop_sh * agent->wickedness,
	     "%7d", is_active_intrvn(i),
	     "%10.2e", (double)agent->lost_good_traffic);

    if (_agents.num < 100) mfprintf(f, 2, "  %-30s", link_str, "%-30s", loc_str);
    fprintf(f, "\n");
  }
  fclose(f);
}

void print_locs() {
  FILE* f = fopen("locs.dat", "w+");
  print_config(&_config, '#', f);
  fprintf(f, "# num %d\n", _locs.num);
  fprintf(f, "# num_occupied %d\n", _locs.num_occupied);
  fprintf(f, "# tot_pop %d\n", _locs.tot_pop);
  fprintf(f, "# occupied_pop %d\n", _locs.occupied_pop);
  fprintf(f, "# min_pop %d\n", _locs.min_pop);
  fprintf(f, "# max_pop %d\n", _locs.max_pop);
  mfprintf(f, 5, "# %10s", "loc", "%10s", "pop", "%10s", "popsh", "%10s", "n agents", "  %-12s\n", "agents");
  for (int i = 0; i < _locs.num; i++) {
    shared loc_t* loc = &_locs._[i];
    double popsh = loc->pop;
    if (loc->num_agents) popsh /= loc->num_agents;
    mfprintf(f, 4, "  %10d", loc->id, "%10d", loc->pop, "%10.1f", popsh, "%10d  ", loc->num_agents);
    int num_items = (loc->num_agents > 15 ? 15 : loc->num_agents);
    for (int a = 0; a < num_items; a++) fprintf(f, "%d,", loc->agent_list[a]);
    fprintf(f, "\n");
  }
  fclose(f);
}

void print_links() {
  FILE* f = fopen("links.dat", "w");
  print_config(&_config, '#', f);
  for (int i = 0; i < _agents.num - 1; i++) {
    for (int j = i + 1; j < _agents.num; j++) {
      if (_agents._[i].links[j]) fprintf(f, "%d %d\n", i, j);
    }
  }
  fclose(f);
}

void print_dists() {
  FILE* f = fopen("dists.dat", "w+");
  print_config(&_config, '#', f);
  fprintf(f, "# av k %.3f\n", get_av_deg());
  fprintf(f, "# mx k %d\n", _agents.max_deg);
  fprintf(f, "# cover %.2f\n", get_locs_covered());
  fprintf(f, "# av sz %.2f\n", get_av_agent_size());
  fprintf(f, "# mx sz %d\n", get_max_agent_size());
  fprintf(f, "# av pop %.2e\n", get_av_agent_pop_sh());

  mfprintf(f, 2, "# %5s", "id", "%10s\n", "av dist");
  for (int i = 0; i < _agents.num; i++) {
    shared agent_t* agent = &_agents._[i];
    mfprintf(f, 2, "  %5d", agent->id, "%10.3f\n", agent->av_dist);
  }
  fclose(f);
}


void save_snapshot() {
  char fname[255];
  sprintf(fname, "snapshot-%d.dat", _agents.num);
  FILE* f = fopen(fname, "w+");
  save_loc_snapshot(f);
  save_agent_snapshot(f);
  fclose(f);
}

void save_loc_snapshot(FILE* f) {
  // save the populations for all locations
  fprintf(f, "%d\n", _locs.num);
  for (int i = 0; i < _locs.num; i++) {
    shared loc_t* loc = get_loc_ptr(i);
    fprintf(f, "%d %d %d ", loc->id, loc->pop,  loc->num_agents);
    for (int ai = 0; ai < loc->num_agents; ai++) fprintf(f, "%d ", loc->agent_list[ai]);
    fprintf(f, "\n");
  }
}

void save_agent_snapshot(FILE* f) {
  fprintf(f, "%d\n", _agents.num);
  for (int i = 0; i < _agents.num; i++) {
    shared agent_t* agent = get_agent_ptr(i, __LINE__);
    fprintf(f, "%d %d %d %d ", agent->id, agent->num_links, agent->birth_time, agent->iters);
    for (int li = 0; li < agent->num_links; li++) fprintf(f, "%d ", agent->links_list[li]);
    fprintf(f, "\n");
  }
}

void restore_snapshot() {
  FILE* f = fopen((char*)_config.snapshot_fname, "r");
  if (!f) FAIL("couldn't open file 'snapshot.dat' to read from\n", "");
  restore_loc_snapshot(f);
  restore_agent_snapshot(f);
  fclose(f);
}

void restore_loc_snapshot(FILE* f) {
  int num;
  if (fscanf(f, "%d\n", &num) != 1) FAIL("could not scan all 6 loc items", "");
  _locs.num = num;
  _locs.num_occupied = 0;
  _locs.tot_pop = 0;
  _locs.occupied_pop = 0;
  _locs.min_pop = MAX_POP;
  _locs.max_pop = 0;
  for (int i = 0; i < _locs.num; i++) {
    shared loc_t* loc = get_loc_ptr(i);
    int id, pop, num_agents;
    if (fscanf(f, "%d %d %d", &id, &pop,  &num_agents) != 3) {
      FAIL("could not scan all 3 loc items", "");
    }
    loc->id = id;
    loc->pop = pop;
    loc->num_agents = num_agents;
    if (num_agents) {
      _locs.num_occupied++;
      _locs.occupied_pop += pop;
    }
    _locs.tot_pop += pop;
    if (pop < _locs.min_pop) _locs.min_pop = pop;
    if (pop > _locs.max_pop) _locs.max_pop = pop;
    for (int ai = 0; ai < loc->num_agents; ai++) {
      int agent_id;
      if (fscanf(f, "%d ", &agent_id) != 1) {
	FAIL("could not scan 1 loc item", "");
      }
      loc->agent_list[ai] = agent_id;
    }
  }
}

void restore_agent_snapshot(FILE* f) {
  int num;
  if (fscanf(f, "%d\n", &num) != 1) FAIL("couldn't read 1 item for agent\n", "");
  _agents.num = num;
  _agents.num_links = 0;
  _agents.max_deg = 0;
  for (int i = 0; i < _agents.num; i++) {
    shared agent_t* agent = get_agent_ptr(i, __LINE__);
    int id, num_links, birth_time, iters;
    if (fscanf(f, "%d %d %d %d ", &id, &num_links, &birth_time, &iters) != 4) {
      FAIL("couldn't read 4 items for agents", "");
    }
    agent->id = id;
    agent->num_links = num_links;
    agent->birth_time = birth_time;
    agent->iters = iters;
    agent->num_locs = 0;
    agent->pop_sh = 0;
    agent->tot_pop = 0;
    _agents.num_links += num_links;
    if (num_links > _agents.max_deg) _agents.max_deg = num_links;
    for (int li = 0; li < agent->num_links; li++) {
      int ai;
      if (fscanf(f, "%d ", &ai) != 1) FAIL("couldn't scan link %d for agent %d\n", li, i);
      agent->links_list[li] = ai;
      agent->links[ai] = 1;
    }
  }
  for (int i = 0; i < _locs.num; i++) {
    shared loc_t* loc = get_loc_ptr(i);
    if (!loc->num_agents) continue;
    double pop_sh = (double)loc->pop / (double)loc->num_agents;
    for (int ai = 0; ai < loc->num_agents; ai++) {
      shared agent_t* agent = get_agent_ptr(loc->agent_list[ai], __LINE__);
      agent->locs[i] = 1;
      agent->locs_list[agent->num_locs++] = i;
      agent->pop_sh += pop_sh;
      agent->tot_pop += _locs._[i].pop;
    }
  }
}

